// 
//  Copyright 2003 OCP-IP
//
// ============================================================================
//      Project : OCP SLD WG, OCP Transaction Level Layer-1
//       Author : Norman Weyrich, Synopsys Inc., weyrich@synopsys.com
//                Anssi Haverinen, Nokia Inc., anssi.haverinen@nokia.com
//                Alan Kamas, www.kamas.com, aok@sonicsinc.com
//         Date : June/06/2003
//
//  Description : Transaction Level - Layer-1 example Slave
//    Features:
//    - Synchronous slave using OCP data class
//    - Non-blocking methods are used
//    - Pipelined Request and Response threads
//    - Single-cycle request with pre-emptive release.
//    - Sideband signal testing
//                
// Parameter    :
//   Template arguments.
//     - TdataCl: Data class containing data members and
//                access methods for the data exchanged between
//                Masters and Slaves
//     - Td: Data type
//     - Ta: Address type
//                
//   Constructor arguments.
//     - sc_module_name: Name of the module instance
//     - ID: Unique number identifying the SLave.
//           ID must be unique among all Slaves attached to the same Bus.
//     - StartAddress: First address of the Slaves memory.
//     - EndAddress: Last address of the Slaves memory.
//                   The address range of all Slaves in a system
//                   must be mutually distinct.
//
//
// ============================================================================

#include "ocp_tl1_slave_sync_sideband3.h"
#include "ocp_tl1_data_cl.h"

// ----------------------------------------------------------------------------
// Process : OCP_TL1_Slave_Sync_Sideband::OCP_TL1_Slave_Sync_Sideband
// 
// Constructor
// ----------------------------------------------------------------------------
template<class TdataCl> OCP_TL1_Slave_Sync_Sideband<TdataCl>::OCP_TL1_Slave_Sync_Sideband
       (
          sc_module_name name_
        , int ID
        , Ta StartAddress
        , Ta EndAddress
        )
          : sc_module          (name_)
          , SlaveP             ("SlavePort")
          , m_ID               (ID)
          , m_StartAddress     (StartAddress)
          , m_EndAddress       (EndAddress)
{
  int i;

  // InitParameters();

  m_NumWordsPerBeat = sizeof(Td);
  m_SmemoryLength = (m_EndAddress - m_StartAddress + 1) / m_NumWordsPerBeat;

  // allcoate memory
  if (m_SmemoryLength > 0)
  {
    try
    {
      m_SlaveMemory = new Td[m_SmemoryLength];
    }
    catch(std::bad_alloc ex)
    {
      cerr << "ERROR: Memory allocation for Slave memory failed "
            << name() << ">" << endl;
      cerr << "       Size  is: " << m_SmemoryLength << endl;
      cerr << "       Error is: " << ex.what() << endl;
      sc_stop();
      return;
    }
  }

  for (i = 0; i < m_SmemoryLength; i++)
    m_SlaveMemory[i] = 0;

  for (i = 0; i < TL_SLAVE_FIFO_DEPTH; i++)
    m_RespEventS[i] = false;

  SC_THREAD(Request);
  sensitive_pos(clk);
  SC_THREAD(Response);
  sensitive << m_RespEvent;
  SC_THREAD(MCmd_VCD);
  SC_METHOD(ReadControl);
  SC_THREAD(SetControlBusySideband);
  sensitive_pos(clk);
}


// ----------------------------------------------------------------------------
// Process : OCP_TL1_Slave_Sync_Sideband::~OCP_TL1_Slave_Sync_Sideband
// 
// Destructor
// ----------------------------------------------------------------------------
template<class TdataCl> OCP_TL1_Slave_Sync_Sideband<TdataCl>::~OCP_TL1_Slave_Sync_Sideband()
{
  delete [] m_SlaveMemory;
}


// ----------------------------------------------------------------------------
//  Method : OCP_TL1_Slave_Sync_Sideband::Request()
//
//  Top level synchronous data sampling method
//
// ----------------------------------------------------------------------------
template<class TdataCl> void OCP_TL1_Slave_Sync_Sideband<TdataCl>::Request()
{
  Td Data;
  Ta Addr;
  OCPMCmdType MCmd;

  int i = 0;
  int ReqIdx = 0;
  int Pause[] = {0, 0, 0, 0, 0, 1, 1, 1, 2, 3};
//  int Pause[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  int cnt = 0;

  while(true)
  {
    wait(); // for rising clock edge

    if (cnt == 0)
    {
      
      
      if (SlaveP->SgetRequestPE()) 
      {
	MCmd = m_DataCl->SgetMCmd(); // check if a request was posted
	if (MCmd != OCP_MCMD_IDLE)
	{
	    m_DataCl->SgetMData(Data);
	    Addr = m_DataCl->SgetMAddr();
	    
	    if (MCmd == OCP_MCMD_RD)
	    {
		// send response in same delta cycle
		m_RespEventS[ReqIdx] = true;
		m_RespEvent.notify();
		
		// get new index (only for those requests, where the
		// response thread is called)
		ReqIdx++;
		if (ReqIdx >= TL_SLAVE_FIFO_DEPTH)
		  ReqIdx = 0;
	    }
	    
#ifdef DEBUG_G1
	    cout << "Slave got request "
		 << "  delta = " << simcontext()->delta_count()
		 << "  time  = " << sc_time_stamp().to_seconds()
		 << "  data  = " << Data << endl;
#endif
	    i++; cnt = Pause[i%10];
	} // end of get request branch
      }
      
      SlaveP->SreleasePE(); // release channel (for next clock cycle)
    }
    else cnt--;

  } // end of while(true) loop
} // end of method

template<class TdataCl> void OCP_TL1_Slave_Sync_Sideband<TdataCl>::MCmd_VCD()
{
  while (true)
  {
    wait(TL_CLK_PERIOD, TL_TIME_UNIT);
    cout << "VCD MCmd " << m_DataCl->MCmd[1-m_DataCl->m_ReqToggle]
         << "  delta = " << simcontext()->delta_count()
         << "  time  = " << sc_time_stamp().to_seconds()
         << endl;
  }
}

// ----------------------------------------------------------------------------
//  Method : OCP_TL1_Slave_Sync_Sideband::Response()
//
//  Top level response method (synchronious)
//
// ----------------------------------------------------------------------------
template<class TdataCl> void OCP_TL1_Slave_Sync_Sideband<TdataCl>::Response()
{
  Td Rdata = 0;
  int ResIdx = 0;

  while(true)
  {
    if (!m_RespEventS[ResIdx])
      wait(); // for m_RespEvent event

    m_DataCl->SputSResp(OCP_SRESP_DVA); // Toggles the response data members
    m_DataCl->SputSData(Rdata++);

    // reset + get new index
    m_RespEventS[ResIdx] = false;
    ResIdx++;
    if (ResIdx >= TL_SLAVE_FIFO_DEPTH)
      ResIdx = 0;

    // send response to Master
    while (!SlaveP->SputResponse())
      wait(clk->posedge_event());
  }
}


// ---------------------------------------------------
//
// Randomly turn ControlBusy on and off
//
// ---------------------------------------------------
template<class TdataCl> void OCP_TL1_Slave_Sync_Sideband<TdataCl>::SetControlBusySideband()
{
 int Pause[] = {0, 2, 5, 10}; // # of cycles between setting sidebands
  int count = 0;
  bool nextControlBusy = false;
  int i=0;

  while(true)
  {
    wait(); // wait for rising edge of clock

    if (count <= 0) {
      // get new Pause value
      count = Pause[i%4]; i++;

      // The wait is over, we get to change ControlBusy
      m_DataCl->CputControlBusy(nextControlBusy);
#ifdef DEBUG_G1
      cout << "Core Slave set ControlBusy: "
           << " delta " << simcontext()->delta_count()
           << " time " << sc_time_stamp().to_seconds()
           << " ControlBusy= " << nextControlBusy << endl;
#endif
      nextControlBusy = !(nextControlBusy);
    } else {
      count--;
    }

  } // end of while loop
}


// ----------------------------------------------------
//
// Wait for a control signal to be written out over the sideband
//
// ----------------------------------------------------
template<class TdataCl> void OCP_TL1_Slave_Sync_Sideband<TdataCl>::ReadControl()
{
  next_trigger( m_DataCl->SidebandControlEvent() );
  // get the current value of Control Sideband

#ifdef DEBUG_G1
  cout << "Core Slave sees new Control signal." << endl;
#endif

  // First - is the new value ready for reading?
  bool controlReady = m_DataCl->CgetControlWr();

  if (!controlReady) {
#ifdef DEBUG_G1
    cout << "Core Slave waiting for ControlWr signal." << endl;
#endif
    next_trigger( m_DataCl->SidebandControlWrEvent() );
  } else {
#ifdef DEBUG_G1
    cout << "Core Slave sees ControlWr signal as true - reading Control." << endl;
#endif
  // print it out
#ifdef DEBUG_G1
    cout << "Slave got Control signal: "
         << "  delta = " << simcontext()->delta_count()
         << "  time  = " << sc_time_stamp().to_seconds()
         << "  Control = " << m_DataCl->CgetControl() << endl;
#endif
  } // end else
}

// ----------------------------------------------------------------------------
//  Method : OCP_TL1_Slave_Sync_Sideband::end_of_elaboration()
//
//  This method is activated at the end of the elaboration phase
//  (when all binding operations are finished).
// ----------------------------------------------------------------------------
template<class TdataCl> void OCP_TL1_Slave_Sync_Sideband<TdataCl>::end_of_elaboration()
{
  sc_module::end_of_elaboration();

  // Initialize debug interface
  SlaveP->SregisterDirectIF(this);

  // Get data structure
  m_DataCl = SlaveP->GetDataCl();

  // Get communication structure
  m_CommCl = SlaveP->GetCommCl();

  // Get system parameter structure
  m_ParamCl = SlaveP->GetParamCl();

  // Put parameter into Channel
  // No parameters needed
}


// ----------------------------------------------------------------------------
//  Method : OCP_TL1_Slave_Sync_Sideband::MPutDirect()
//
//  Debug interface method.
//  Read/Write data, Master to Slave direction
//  Returns true for success and false for failure
//
// ----------------------------------------------------------------------------
template<class TdataCl> bool OCP_TL1_Slave_Sync_Sideband<TdataCl>::MputDirect(
         int MasterID, bool IsWrite, Td *Data, Ta Address, int NumWords)
{
  int i;
  int NumBeats = NumWords / m_NumWordsPerBeat;
  int Offset = (Address - m_StartAddress) / m_NumWordsPerBeat;

  if ((Offset >= 0) && ((Offset + NumBeats) <= m_SmemoryLength))
  {
    if (IsWrite)
    {
      // write request
      for (i = 0; i < NumBeats; i++)
        m_SlaveMemory[Offset + i] = Data[i];
    }
    else
    {
      // read request
      for (i = 0; i < NumBeats; i++)
        Data[i] = m_SlaveMemory[Offset + i];
    }
    return(true);
  }
  else
  {
    cout << "Error: Address out of range at Slave " << m_ID << endl
         << "       Address=" << Address << "  StartAddr=" << m_StartAddress
         << "  _EndAddr=" << m_EndAddress
         << "  NumWords=" << NumWords
         << endl;
    return(false);
  }
}


// ----------------------------------------------------------------------------
//
//  Instantiation of the Slave
//
// ----------------------------------------------------------------------------
template class OCP_TL1_Slave_Sync_Sideband<TL1_TEMPL_DATA_CL >; // see ocp_tl1_globals.h
