///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// (c) Copyright OCP-IP 2003, 2004
// OCP-IP Confidential and Proprietary
//
//
//============================================================================
//      Project : OCP SLD WG
//       Author : Alan Kamas, for Sonics Inc.
//                Anssi Haverinen, Nokia Inc.
//                Joe Chou, Sonics Inc.
//                
//         Date: 05/21/2004
//
//  Description :  OCP TL 1 SystemC Channel Model
//        This version of Channel is derived from the original 
//        OCP SLD WG, OCP Transaction Level Layer-1 model by
//        Norman Weyrich, Anssi Haverinen, and Joe Chou.
//        Please see that model for details.
//
// This version adds full channel access through the ports;
// single function calls to send and receive requests, data, and responses;
// and clocking to resolve OCP timing.
//
// Change History:
//
// 12/14/2003 Original revision
// 05/21/2004 Adds clocked blocking calls, moves
//            OCP trace monitor to a separate file, makes OCP
//            checker optional, adds *accept(bool)
//            calls for accepting transaction before it is issued.
//            Removes reset event sensitivity from blocking calls,
//            relying on new tl_channel reset functionality.
//            Data and address mask disabled to support array data types
//            Constructors modified.
// 
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef _OCP_TL1_CHANNEL_H
#define _OCP_TL1_CHANNEL_H


#include <math.h>
#include <stdio.h>

#include <fstream>

#include <string>
#include <map>

#include "systemc.h"

#include "tl_channel.h"
#include "ocp_globals.h"
#include "ocp_tl1_master_if.h"
#include "ocp_tl1_slave_if.h"
#include "ocp_mon.h"

// Added for gcc 3.3 Compatibility
using namespace std;

// -----------------------------------------------------------------
// -----------------------------------------------------------------

template<typename TdataCl>
class OCP_TL1_Channel :
  public TL_Channel<TdataCl>,
  public OCP_TL1_MasterIF<TdataCl>,
  public OCP_TL1_SlaveIF<TdataCl>
{
public:
  // -----------------------------------------------------------------
  // public types
  // -----------------------------------------------------------------

  // Give OCP monitor an access to protected data
  friend class OCPMon<TdataCl>;

  typedef typename TdataCl::DataType Td;
  typedef typename TdataCl::AddrType Ta;

  // Needed for when the ocp clock tick process is used
  SC_HAS_PROCESS(OCP_TL1_Channel);


  //---------------------------------------------------------------
  // constructor for clocked and non-clocked channels
  //---------------------------------------------------------------

  OCP_TL1_Channel(std::string name,
                  sc_in_clk * clk,
		  std::string monFileName = ""
                  )
    : TL_Channel<TdataCl>(name.c_str(),true,true,true,false),
    m_clkPeriod(1),
    m_clkTimeUnit(SC_NS),
    m_addressMask(0x3),
    m_dataMask(0xFFFFFFFF),
    m_exitAfterOCPMon(false),
    m_clocked(true),
    m_self_timed(false),
    m_clk(NULL),
    m_in_clk(clk),
    m_sig_clk(NULL),
    m_monitor(NULL)
  {

#ifndef NDEBUG      
    m_doRunCheck = true;
#else
    m_doRunCheck = false;
#endif
    if (clk==NULL)
      m_clocked = false;
    else {
      if (! monFileName.empty() ) {
	std::string n = (std::string)name.c_str()+"mon";
        m_monitor = new OCPMon< OCP_TL1_DataCl<Td, Ta> > (n.c_str(), (OCP_TL1_Channel<TdataCl> *) this, (std::string) monFileName, (sc_in_clk *)clk, false);
      }
    }

  }

  //---------------------------------------------------------------
  // constructor for clocked and non-clocked channels
  //---------------------------------------------------------------

  OCP_TL1_Channel(std::string name,
                  sc_clock * clk,
		  std::string monFileName = ""
                  )
    : TL_Channel<TdataCl>(name.c_str(),true,true,true,false),
    m_clkPeriod(1),
    m_clkTimeUnit(SC_NS),
    m_addressMask(0x3),
    m_dataMask(0xFFFFFFFF),
    m_exitAfterOCPMon(false),
    m_clocked(true),
    m_self_timed(false),
    m_clk(clk),
    m_in_clk(NULL),
    m_sig_clk(NULL),
    m_monitor(NULL)
  {

#ifndef NDEBUG      
    m_doRunCheck = true;
#else
    m_doRunCheck = false;
#endif
    if (clk==NULL)
      m_clocked = false;
    else {
      if (! monFileName.empty() ) {
	std::string n = (std::string)name.c_str()+"mon";
        m_monitor = new OCPMon< OCP_TL1_DataCl<Td, Ta> > (n.c_str(), (OCP_TL1_Channel<TdataCl> *) this, (std::string) monFileName, (sc_clock *)clk, false);
      }
    }

  }

  //---------------------------------------------------------------
  // constructor for clocked channels
  //---------------------------------------------------------------

  OCP_TL1_Channel(std::string name,
                  sc_signal<bool> * clk,
		  std::string monFileName = ""
                  )
    : TL_Channel<TdataCl>(name.c_str(),true,true,true,false),
    m_clkPeriod(1),
    m_clkTimeUnit(SC_NS),
    m_addressMask(0x3),
    m_dataMask(0xFFFFFFFF),
    m_exitAfterOCPMon(false),
    m_clocked(true),
    m_self_timed(false),
    m_clk(NULL),
    m_in_clk(NULL),
    m_sig_clk(clk),
    m_monitor(NULL)
  {

#ifndef NDEBUG      
    m_doRunCheck = true;
#else
    m_doRunCheck = false;
#endif
    if (clk==NULL)
      m_clocked = false;
    else {
      if (! monFileName.empty() ) {
	std::string n = (std::string)name+"mon";
        m_monitor = new OCPMon< OCP_TL1_DataCl<Td, Ta> > (n.c_str(), (OCP_TL1_Channel<TdataCl> *) this, (std::string) monFileName, (sc_signal<bool> *)clk, false);
      }
    }

  }

  //---------------------------------------------------------------
  // simple constructor for self-timed channels
  //---------------------------------------------------------------

  OCP_TL1_Channel(std::string name,
                  double clock_period,
                  sc_time_unit clock_time_unit = SC_NS,
		  std::string monFileName = ""
                  )
    : TL_Channel<TdataCl>(name.c_str(),true,true,true,(bool)clock_period),
      m_clkPeriod(clock_period),
      m_clkTimeUnit(clock_time_unit),
      m_addressMask(0x3),
      m_dataMask(0xFFFFFFFF),
      m_exitAfterOCPMon(false),
      m_clocked(false),
      m_self_timed((bool)clock_period),
      m_clk(NULL),
      m_in_clk(NULL),
      m_sig_clk(NULL),
      m_monitor(NULL)
  {

#ifndef NDEBUG      
    m_doRunCheck = true;
#else
    m_doRunCheck = false;
#endif    

    if (! monFileName.empty() ) {
      std::string n = (std::string)name+"mon";
      m_monitor = new OCPMon< OCP_TL1_DataCl<Td, Ta> > (n.c_str(), (OCP_TL1_Channel<TdataCl> *) this, clock_period, clock_time_unit, (std::string) monFileName, false);
    }
  }

  //---------------------------------------------------------------
  // constructor for self-timed and non-clocked channels
  //---------------------------------------------------------------
  OCP_TL1_Channel(std::string name,
                  bool sync = true,
                  bool use_event = true,
                  bool use_default_event = true,
                  sc_trace_file* vcd_tf = NULL,
                  double clock_period = 0,
                  sc_time_unit clock_time_unit = SC_NS,
		  std::string monFileName = "",
                  bool runtimeCheck = true
                  )
    : TL_Channel<TdataCl>(name.c_str(),sync,use_event,use_default_event, (bool)clock_period),
      m_clkPeriod(clock_period),
      m_clkTimeUnit(clock_time_unit),
      m_doRunCheck(runtimeCheck),
      m_addressMask(0x3),
      m_dataMask(0xFFFFFFFF),
      m_exitAfterOCPMon(false),
      m_clocked(false),
      m_self_timed((bool)clock_period),
      m_clk(NULL),
      m_in_clk(NULL),
      m_sig_clk(NULL),
      m_monitor(NULL)
  {

    // The OCP Specific TL1 channel only accepts the following:
    assert(sync);
    if (m_self_timed) {
      assert(use_event);
      assert(use_default_event);
    }
    
    // The OCP TL1 channel no longer accept does trace dumps.
    // This functionality was replaced with the OCP Monitor output.
    // The vcd file name parameter is kept for backward compatibility
    assert( vcd_tf == NULL ); 

    if (! monFileName.empty() ) {
      std::string n = (std::string)name+"mon";
      m_monitor = new OCPMon< OCP_TL1_DataCl<Td, Ta> > (n.c_str(), (OCP_TL1_Channel<TdataCl> *) this, clock_period, clock_time_unit, (std::string) monFileName, false);
    }
  }

  //---------------------------------------------------------------
  // destructor
  //---------------------------------------------------------------
  virtual ~OCP_TL1_Channel()
  {
  }

public:

  // -----------------------------------------------------------------
  // Basic Channel Methods
  // -----------------------------------------------------------------
  virtual void register_port(sc_port_base& port, const char* if_typename)
  { TL_Channel<TdataCl>::register_port(port,if_typename); }

  /////////////////////////////////////////
  // OCP Specific Commands
  /////////////////////////////////////////

  //---------------------------------------------------------------------
  // Configuration Functions
  //---------------------------------------------------------------------

  void 
  setConfiguration( MapStringType& passedMap )
  {
    // Use the passed data base to set up the configuration class
    // NOTE: no "prefix" needed, "name()" passed for debugging
    m_ParamCl->setOCPConfiguration( name(), passedMap );
    
    // Set up the address mask.
    m_addressMask=0;
/*      for (int i=0; i<(m_ParamCl->addr_wdth); i++) { */
/*        m_addressMask = m_addressMask << 1; */
/*        m_addressMask = m_addressMask +1; */
/*      } */
    
    // Set up the data mask.
    m_dataMask=0;
/*      for (int i=0; i<(m_ParamCl->data_wdth); i++) { */
/*        m_dataMask = m_dataMask << 1; */
/*        m_dataMask = m_dataMask +1; */
/*      } */
    

    // Do OCP monitor startup preparations if installed
    if (m_monitor!=NULL) {
      m_monitor->startup();
      if (m_doRunCheck)
	m_monitor->runcheck();
    }
  }


  //---------------------------------------------------------------------
  // Timing Functions
  //---------------------------------------------------------------------
  //
  void ocpWait(double count = 1) {
    if (m_clocked){
      for (int i=0;i<count;i++)
	if (m_clk!=NULL)
	  wait(m_clk->posedge_event()); // For clocked channel
	else if (m_in_clk!=NULL)
	  wait(m_in_clk->posedge_event()); // For clocked channel
	else if (m_sig_clk!=NULL)
	  wait(m_sig_clk->posedge_event()); // For clocked channel
	else {
	  fatalError("ocpWait()", "Only self-timed or clocked channels can use ocpWait");
	}
    }
    else if (m_self_timed)
      wait(count*m_clkPeriod, m_clkTimeUnit); // For self-timed channel
    else {
      fatalError("ocpWait()", "Only self-timed or clocked channels can use ocpWait");
    }
  }

  // Handle the case of reset during a longer wait
  // true if successful wait. false if interrupted by reset.
  // Very slow call, be aware.
  bool ocpSafeWait(double count = 1) {
    if (get_reset()) {
      return false;
    }
    if (m_clocked){
      for (int i=0;i<count;i++)
	if (m_clk!=NULL)
	  wait(m_clk->posedge_event()|m_CommCl->ResetStartEvent);
	else if (m_in_clk!=NULL)
	  wait(m_in_clk->posedge_event()|m_CommCl->ResetStartEvent);
	else if (m_sig_clk!=NULL)
	  wait(m_sig_clk->posedge_event()|m_CommCl->ResetStartEvent);
	else {
	  return false;
	}
    }
    else if (m_self_timed)
      wait(count*m_clkPeriod, m_clkTimeUnit , m_CommCl->ResetStartEvent);
    else
      return false;
    return (! (get_reset()) );
  }

  //////////////////////////////////////////////////////////////
  // OCP TL1 Methods
  //////////////////////////////////////////////////////////////

  // -----------------------------------------------------------------
  // OCP TL1 commands with overloaded parameters
  // -----------------------------------------------------------------
  bool getOCPResponse( OCPResponseGrp<Td>& myResponse )
  { return getOCPResponse(myResponse, false); }
  bool getOCPResponseBlocking( OCPResponseGrp<Td>& myResponse )
  { return getOCPResponseBlocking(myResponse, false); }
  bool getOCPRequest( OCPRequestGrp<Td,Ta>& myRequest )
  { return getOCPRequest(myRequest, false); }
  bool getOCPRequestBlocking( OCPRequestGrp<Td,Ta>& myRequest )
  { return getOCPRequestBlocking(myRequest, false); }
  bool getOCPDataHS( OCPDataHSGrp<Td>& myData )
  { return getOCPDataHS(myData, false); }
  bool getOCPDataHSBlocking( OCPDataHSGrp<Td>& myData )
  { return getOCPDataHSBlocking(myData, false); }


  //---------------------------------------------------------------------
  // public methods for the OCP Request phase
  //---------------------------------------------------------------------

  bool getSBusy(void) const 
  { 
    if (m_self_timed)
      return ( (TL_Channel<TdataCl>::MgetSbusy()) || (sc_simulation_time() == m_CommCl->RequestEndTime) ); 
    else
      return (TL_Channel<TdataCl>::MgetSbusy()); 
  }

  bool 
  startOCPRequest(OCPRequestGrp<Td,Ta>& newRequest)
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    // Is the thread free?
    if (m_ParamCl->sthreadbusy_exact)
    {
      // SThreadBusy is part of the channel and we are required to check it.
      // Now we need to test to see if the thread of the request is busy.
      if ( m_DataCl->isThisThreadBusy(newRequest.MThreadID, getSThreadBusy()) )
      {
	// The thread is busy. Fail.
// #ifndef NDEBUG              
	//               cout << name() << ": startOCPRequest failed. It was called for a busy thread on a channel with sthreadbusy_exact." << endl;
// #endif
	return false;
      }
    }
      
    // Has a request already been accepted this cycle? Only one request per ocp cycle is allowed.
    if (m_self_timed)
      if ( sc_simulation_time() == m_CommCl->RequestEndTime ) {
	return false;
      }

    // Mark that a new request is going on the channel
    if ( !(MputRequest()) )
    {
      return false;
    }
    
    // Removed from this release
    // Make sure the data & address fits the channel by masking off extra bits.
    //    newRequest.MData &= m_dataMask;
    //    newRequest.MAddr &= m_addressMask;

    // Load the new request info onto the channel
    m_DataCl->MputRequestGroup(newRequest);

    return true;
  }
  
  bool startOCPRequestBlocking(OCPRequestGrp<Td,Ta>& newRequest)
  { 
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    } 
    bool tmp = true;
    if (m_clocked) // Clocked channel
    {
      while (tmp)
      {
	tmp = getSBusy();
	ocpWait();
      }
      startOCPRequest(newRequest);
    }
    else if (m_self_timed) // Self-timed channel
    {
      if (m_CommCl->BlockingRequestPending) 
      {
	// Blocking request already started - abort.
	return false;
      }
      
      // Mark that there is a Blocking Request in progress
      m_CommCl->BlockingRequestPending = true;
      
      // Is there a request in progress right now?
      if (m_CommCl->RequestPending || m_CommCl->RequestStart) 
      {
	// There is a request in progress
	// We must wait until it has finished
	wait(m_CommCl->RequestEndEvent);
	
	// Was the channel reset?
	if (get_reset()) {
	  // In reset. Give up.
	  return false;
	}
	
	// At this point the previous command has been accepted.
	// We now need to wait for the next clock cycle to start a new request.
	ocpWait();
      }
      
      // The previous request (if any) is over. It's our turn now.
      // Try to send a request
      m_CommCl->BlockingRequestPending = false;
      while ( !startOCPRequest(newRequest) )
      {
	// Was the channel reset?
	if (get_reset()) {
	  // In reset. Give up.
	  return false;
	}
	// The request failed. 
	// Lock up the channel, wait, and try again.
	m_CommCl->BlockingRequestPending = true;
	ocpWait();
	m_CommCl->BlockingRequestPending = false;
      }
    }
    else
    {
      fatalError("startOCPRequestBlocking()", "Only self-timed or clocked channels can use blocking calls");

      return false;
    }
    // We are done
    return true;
  }

  bool 
  getSCmdAccept(void) const 
  { 
    //      return ((!m_CommCl->RequestPending) && (m_CommCl->RequestAccepted 
    //					      || m_CommCl->RequestEarlyEnd) );
    if (!m_ParamCl->cmdaccept)
      return(true);
    else
      return (!getSBusy());
  }

  void
  waitSCmdAccept(void)
  {
    // is the channel in reset?
    if (!get_reset()) {
      if (!getSBusy() || !m_ParamCl->cmdaccept){
	// The command has already been accepted
	return;
      } 
      wait(m_CommCl->RequestEndEvent);
    }
  }

  void
  waitSDataAccept(void)
  {
    // is the channel in reset?
    if (!get_reset()) {
      if (!getSBusyDataHS() || !m_ParamCl->dataaccept) {
        // The data has already been accepted
        return;
      } 
      wait(m_CommCl->DataRequestEndEvent);
    }
  }

  void
  waitMRespAccept(void)
  {
    // is the channel in reset?
    if (!get_reset()) {
      if (!getMBusy() || !m_ParamCl->respaccept) {
        // The command has already been accepted
        return;
      } 
      wait(m_CommCl->ResponseEndEvent);
    }
  }

  unsigned int 
  getSThreadBusy(void) const 
  { 
    // is the channel in reset?
    if (m_CommCl->Reset) {
      // In reset. Give up.
      return false;
    }
    // The signal sthreadbusy must be part of the channel
    // for this function to make sense.
    assert( m_ParamCl->sthreadbusy );
    
    return m_DataCl->MgetSThreadBusy();
  }

  void 
  putSThreadBusy(unsigned int sthreadbusy) 
  { 
    // is the channel in reset?
    if (!get_reset()) {
      // The signal sthreadbusy must be part of the channel
      // for this function to make sense.
      assert( m_ParamCl->sthreadbusy );
      
      m_DataCl->SputSThreadBusy( sthreadbusy );
    }
  }

  void 
  putNextSThreadBusy(unsigned int sthreadbusy) 
  { 
    // is the channel in reset?
    if (!get_reset()) {
      // The signal sthreadbusy must be part of the channel
      // for this function to make sense.
      assert( m_ParamCl->sthreadbusy );
      
      m_DataCl->SputNextSThreadBusy( sthreadbusy );
    }
  }
  
  // NOTE: acceptRequest should default to false
  bool getOCPRequest(OCPRequestGrp<Td,Ta>& myRequest, bool acceptRequest) 
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    if ( !SgetRequest(acceptRequest) ){ 
      // Could not get a request. Maybe there is not one available.
      return false;
    }
    
    // We got the request - now get the request info for the user
    myRequest = m_DataCl->SgetRequestGroup();
    m_DataCl->SgetMCmd(); 
    return true;
  }

  // NOTE: acceptRequest should default to false
  bool getOCPRequestBlocking(OCPRequestGrp<Td,Ta>& myRequest, bool acceptRequest) 
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    // Is there a blocking get request already in progress? 
    // If so, that getRequestBlocking command has priority to read first.
    if ( m_CommCl->BlockingGetRequestPending ) {
      return false;
    }
    
    // We are the blocking getRequest. Mark it so that no one else can go.
    m_CommCl->BlockingGetRequestPending = true;
    
    // Is there a request for us to read?
    if ( !(m_CommCl->RequestStart) || !(m_CommCl->RequestUnread) ) {
      // No request yet. Must be patient.
      wait(m_CommCl->RequestStartEvent);
      if (get_reset()) {
	// The channel has been reset. Give up.
	return false;
      }
    }

    // There should be a request now
    assert( m_CommCl->RequestStart == true );
    assert( m_CommCl->RequestUnread == true );

    // Blocking is over - get the request normally
    m_CommCl->BlockingGetRequestPending = false;
    // get and return the request
    return getOCPRequest(myRequest, acceptRequest);
  }

  bool 
  putSCmdAccept() 
  { 
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    // Was this command called at the correct time?
    if (! m_CommCl->RequestStart ) {
#ifndef NDEBUG
      cout << name() << ": WARNING - putSCmdAccept called when there was no request on the channel." << endl;
#endif
      return false;
      }
    Srelease();
    return true;
  }

  void
  putSCmdAccept(bool scmdaccept) 
  { 
    // is the channel in reset?
    if (!get_reset()) {
      if (scmdaccept)
	SreleasePE();
      else
	SunreleasePE();
    }
  }

  const sc_event& RequestStartEvent(void) const 
  { return (m_CommCl->RequestStartEvent); }

  const sc_event& RequestEndEvent(void) const 
  { return (m_CommCl->RequestEndEvent); }

  //---------------------------------------------------------------------
  // public methods for the OCP Response Phase
  //---------------------------------------------------------------------


  bool getMBusy(void) const 
  { 
    if (m_self_timed)
      return ( (TL_Channel<TdataCl>::SgetMbusy()) || (sc_simulation_time() == m_CommCl->ResponseEndTime) );
    else
      return (TL_Channel<TdataCl>::SgetMbusy());
  }
  
  bool startOCPResponse(OCPResponseGrp<Td>& newResponse)
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    // Is the thread free?
    if (m_ParamCl->mthreadbusy_exact)  {
      // MThreadBusy is part of the channel and we are required to check it.
      // Now we need to test to see if the thread is busy.
      if ( m_DataCl->isThisThreadBusy(newResponse.SThreadID, getMThreadBusy()) ){
	// The thread is busy. Fail.
#ifdef DEBUG_C
	cout << name() << ": startOCPResponse failed. It was called for a busy thread on a channel with mthreadbusy_exact." << endl;
#endif
	return false;
      }
    }

#ifndef NDEBUG
    // Is the thread free?
    if (m_ParamCl->mthreadbusy)
      {
	// MThreadBusy is part of the channel. 
	// Now we test to see if the thread of the request is busy.
	if ( m_DataCl->isThisThreadBusy(newResponse.SThreadID, getMThreadBusy()) ) {
	  // The thread is busy. Warn the user.
	  cout << name() << ": WARNING - startOCPResponse called for a busy thread." << endl;
	}
      }
#endif
    
      // Has a response already been accepted this cycle? Only one response per ocp cycle is allowed.
    if (m_self_timed)
      if ( sc_simulation_time() == m_CommCl->ResponseEndTime ) {
	return false;
      }
    
    // Mark that a new response is going on the channel
    if ( !(SputResponse()) ) {
      // Could not start the response
      return false;
    }
    
    // #ifndef NDEBUG
    //    cout << name() << ": startOCPResponse placed the new response on the channel." << endl;
    // #endif
    
    // Removed from this release
    // Make sure the data fits on the channel    
    //    newResponse.SData &= m_dataMask;
    
    // The response is ours to send. Put our data unto the channel
    m_DataCl->SputResponseGroup(newResponse);
    
    // All done
    return true;
  }
  
  bool 
  startOCPResponseBlocking(OCPResponseGrp<Td>& newResponse)
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    bool tmp = false;
    if (m_clocked) // Clocked channel
    {
      while (!tmp)
      {
	ocpWait();
	tmp = startOCPResponse(newResponse);
      }
    }
    else if (m_self_timed)// Self-timed channel
    {
      if (m_CommCl->BlockingResponsePending) {
          // Blocking response already started - abort.
          return false;
      }

      // Mark that there is a Blocking Response in progress
      m_CommCl->BlockingResponsePending = true;

      // Is there a response in progress right now?
      if (m_CommCl->ResponsePending || m_CommCl->ResponseStart) {
          // There is a response in progress
          // We must wait until it has finished
// #ifndef NDEBUG
// cout << name() << ": startOCPResponseBlocking is waiting for previous response to end." << endl;
// #endif
          wait(m_CommCl->ResponseEndEvent);

          if (get_reset()) {
              // Channel was reset while we were waiting. Give up.
              return false;
          }

          // At this point the previous response has been accepted.
          // We now need to wait for the next clock cycle to start a new response
// #ifndef NDEBUG
// cout << name() << ": startOCPResponseBlocking has detected that the previous response has been accepted" << endl;
// #endif
          ocpWait();
      }

      // The previous response (if any) is over. It's our turn now.
      // Try to send a response
      m_CommCl->BlockingResponsePending = false;
      while (!startOCPResponse(newResponse)) {
          // The response failed.

          // Was failure do to a channel reset?
          if (get_reset()) {
              // Channel was reset while we were waiting. Give up.
              return false;
          }

          // Lock up the channel, wait and try again.
// #ifndef NDEBUG
// cout << name() << ": startOCPResponseBlocking has failed to put response on the empty channel. Will wait and try again." << endl;
// #endif
          m_CommCl->BlockingResponsePending = true;
          ocpWait();
          m_CommCl->BlockingResponsePending = false;
      }
    }
    else
    {
      fatalError("startOCPResponseBlocking()", "Only self-timed or clocked channels can use blocking calls");

      return false;
    }
    // we are done
    return true;
  }

  bool 
  getMRespAccept(void) const 
  { 
    //      return ( !(m_CommCl->ResponsePending) && (m_CommCl->ResponseAccepted
    //					|| m_CommCl->ResponseEarlyEnd) );
    if (!m_ParamCl->respaccept) {
      return(true);
    }
    else {
      return(!getMBusy());
    }
  }


  unsigned int getMThreadBusy(void) const
  { 
    // is the channel in reset?
    if (m_CommCl->Reset) {
      // In reset. Give up.
      return false;
    }
    // There is trouble here if this function is being called on a channel
    // that does not have mthreadbusy signal.
    assert( m_ParamCl->mthreadbusy == true);
    
    return m_DataCl->SgetMThreadBusy();
  }

  void 
  putMThreadBusy(unsigned int mthreadbusy) 
  { 
    // is the channel in reset?
    if (!get_reset()) {
      // There is trouble here if this function is being called on a channel
      // that does not have mthreadbusy signal.
      assert( m_ParamCl->mthreadbusy == true);
      
      m_DataCl->MputMThreadBusy( mthreadbusy );
    }
  }

  void 
  putNextMThreadBusy(unsigned int mthreadbusy) 
  { 
    // is the channel in reset?
    if (!get_reset()) {
      // There is trouble here if this function is being called on a channel
      // that does not have mthreadbusy signal.
      assert( m_ParamCl->mthreadbusy == true);
      
      m_DataCl->MputNextMThreadBusy( mthreadbusy );
    }
  }


  // NOTE: acceptResponse should default to false
  bool 
  getOCPResponse(OCPResponseGrp<Td>& newResp, bool acceptResponse)
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }   
    if ( !MgetResponse(acceptResponse) )  {
      // Could not get a response. Maybe there is not one available.
      return false;
    }

    // We got a response: now get the data out of the channel
    newResp = m_DataCl->MgetResponseGroup();
    m_DataCl->MgetSResp();
    return true;
  }

  // NOTE: acceptResponse should default to false
  bool 
  getOCPResponseBlocking(OCPResponseGrp<Td>& newResp, bool acceptResponse)
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    // Is there a blocking get response already in progress? 
    // If so, that getResponseBlocking command has priority to read first.
    if ( m_CommCl->BlockingGetResponsePending ) {
      return false;
    }
    
    // We are the blocking getResponse. Mark it so that no one else can go.
    m_CommCl->BlockingGetResponsePending = true;
    
    // Is there a response for us to read?
    if ( !(m_CommCl->ResponseStart) || !(m_CommCl->ResponseUnread) ) {
      // No response yet. Must be patient.
      wait(m_CommCl->ResponseStartEvent);
      // Did the channel get reset while we were waiting?
      if (get_reset()) {
	// The channel has been reset. Give up.
	return false;
      }
    }
    
    // There should be a response now
    assert( m_CommCl->ResponseStart == true );
    assert( m_CommCl->ResponseUnread == true );
    
    // Blocking is over - get the response normally
    m_CommCl->BlockingGetResponsePending = false;
    // get and return the response
    return getOCPResponse(newResp, acceptResponse);
  }

  bool putMRespAccept() 
  { 
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    // Was this command called at the correct time?
    if (! m_CommCl->ResponseStart )
      {
#ifndef NDEBUG
	cout << name() << ": WARNING - putMRespAccept called when there was no response on the channel." << endl;
#endif
	return false;
      }
    Mrelease();
    return true;
  }
      
  void
  putMRespAccept(bool mrespaccept) 
  {
    // is the channel in reset?
    if (!get_reset()) {
      if (mrespaccept)
	MreleasePE();
      else
	MunreleasePE();
    }
  }

  const sc_event& ResponseStartEvent(void) const 
  { return (m_CommCl->ResponseStartEvent); }

  const sc_event& ResponseEndEvent(void) const 
  { return (m_CommCl->ResponseEndEvent); }

  //---------------------------------------------------------------------
  // public methods for the OCP Data Handshake Phase
  //---------------------------------------------------------------------

  bool getSBusyDataHS(void) const 
  { 
    if (m_self_timed)
      return ( (TL_Channel<TdataCl>::MgetSbusyData())|| (sc_simulation_time() == m_CommCl->DataRequestEndTime) ); 
    else
      return (TL_Channel<TdataCl>::MgetSbusyData()); 
  }

  bool
  startOCPDataHS(OCPDataHSGrp<Td>& datahs)
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    // Is the thread free?
    if (m_ParamCl->sdatathreadbusy_exact) {
      // SDataThreadBusy is part of the channel and we are required to check it.
      // Now we need to test to see if the thread of the data handshake is busy.
      if ( m_DataCl->isThisThreadBusy(datahs.MDataThreadID, getSDataThreadBusy()) ) {
	// The thread is busy. Fail.
	// #ifndef NDEBUG
	// cout << name() << ": putData failed. It was called for a busy thread on a channel with sdatathreadbusy_exact." << endl;
	// #endif
	return false;
      }
    }
    // Has a data handshake already been accepted this cycle? Only one data handshake per ocp cycle is allowed.
    if (m_self_timed)
      if ( sc_simulation_time() == m_CommCl->DataRequestEndTime ) {
	return false;
      }

    // Mark that new data is going on the channel
    if ( !(MputDataRequest()) ){
      return false;
    }
    // Removed  from this release
    // Make sure the data fits on the channel    
    //    datahs.MData &= m_dataMask;
    
    // Load the new request info onto the channel
    m_DataCl->MputDataHSGroup(datahs);
    
    return true;
  }

  bool 
  startOCPDataHSBlocking(OCPDataHSGrp<Td>& newData)
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    bool tmp = false;
    if (m_clocked) // Clocked channel
    {
      while (!tmp)
      {
	ocpWait();
	tmp = startOCPDataHS(newData);
      }
    }
    else if (m_self_timed)// Self-timed channel
    {
      if (m_CommCl->BlockingDataRequestPending) 
      {
          // Blocking data request already started - abort.
          return false;
      }
    
      // Mark that there is a Blocking Data Request in progress
      m_CommCl->BlockingDataRequestPending = true;
    
      // Is there a data request in progress right now?
      if (m_CommCl->DataRequestPending || m_CommCl->DataRequestStart) 
      {
          // There is a data request in progress
          // We must wait until it has finished
// #ifndef NDEBUG
// cout << name() << ": startOCPDataHSBlocking is waiting for previous data request to end." << endl;
// #endif
          wait(m_CommCl->DataRequestEndEvent);

          // Did the channel enter reset while we were waiting?
          if (get_reset()) {
              // The channel has been reset. Give up.
              return false;
          }

          // At this point the previous data hand shake has been accepted.
          // We now need to wait for the next clock cycle to start new data.
          ocpWait();
      }

      // The previous data hand shake (if any) is over. It's our turn now.
      // Try to send data
      m_CommCl->BlockingDataRequestPending = false;
      while ( !startOCPDataHS(newData) )
      {
          // The data request failed. 

          // Did the channel enter reset while we were waiting?
          if (get_reset()) {
              // The channel has been reset. Give up.
              return false;
          }

          // Lock up the channel, wait, and try again.
          // NOTE: failure could be due to thread busy. Are we sure we want to keep trying?
          m_CommCl->BlockingDataRequestPending = true;
          ocpWait();
          m_CommCl->BlockingDataRequestPending = false;
      }
    }
    else
    {
      fatalError("startOCPDataHSBlocking()", "Only self-timed or clocked channels can use blocking calls");
      return false;
    }
    // We are done
    return true;
  }

  // NOTE: acceptData should default to false
  bool getOCPDataHS(OCPDataHSGrp<Td>& myDataHS, bool acceptData)
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    if ( !SgetDataRequest(acceptData) ){ 
      // Could not get data. Maybe it is not available.
      return false;
    }
    
    // We got the data - now get the data hand shake info for the user
    myDataHS = m_DataCl->SgetDataHSGroup();
    m_DataCl->SgetMDataValid();
    return true;
  }

  // NOTE: acceptData should default to false
  bool getOCPDataHSBlocking(OCPDataHSGrp<Td>& myDataHS, bool acceptData)
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    }
    // Is there a blocking get data already in progress? 
    // If so, that getDataRequestBlocking command has priority to read first.
    if ( m_CommCl->BlockingGetDataRequestPending ) {
      return false;
    }

    // We are the blocking getData. Mark it so that no one else can go.
    m_CommCl->BlockingGetDataRequestPending = true;
    
    // Is there data for us to read?
    if ( !(m_CommCl->DataRequestStart) || !(m_CommCl->DataRequestUnread) ) {
      // No data yet. Must be patient.
      wait(m_CommCl->DataRequestStartEvent);
      // Did the channel enter reset while we were waiting?
      if (get_reset()) {
	// The channel has been reset. Give up.
	return false;
      }
    }

    // There should be data now
    assert( m_CommCl->DataRequestStart == true );
    assert( m_CommCl->DataRequestUnread == true );
    
    // Blocking is over - get the data normally
    m_CommCl->BlockingGetDataRequestPending = false;
    // get and return the data
    return getOCPDataHS(myDataHS, acceptData);
  }

  bool putSDataAccept() 
  {
    // is the channel in reset?
    if (get_reset()) {
      // In reset. Give up.
      return false;
    } 
    // Was this command called at the correct time?
    if (! m_CommCl->DataRequestStart ) {
#ifndef NDEBUG
      cout << name() << ": WARNING - putSDataAccept called when there was no data on the channel." << endl;
#endif
      return false;
    }
    SreleaseData();
    return true;
  }

  void
  putSDataAccept(bool dataaccept) 
  {
    // is the channel in reset?
    if (!get_reset()) {
      if (dataaccept)
	SreleaseDataPE();
      else
	SunreleaseDataPE();
    }
  }

  bool getSDataAccept(void) const 
  { 
    //      return ( !(m_CommCl->DataRequestPending) && (m_CommCl->DataRequestAccepted 
    //				    ||m_CommCl->DataRequestEarlyEnd) ) ;
    if (!m_ParamCl->dataaccept) {
      return (true);
    }
    else {
      return (!getSBusyDataHS());
    }
  }

  // Optional command provided to the verify the timing of the Master
  // NOTE: doesn't actually do anything
  // NOTE: Not required.
  void putNullMDataValid(void) const
  { 
#ifndef NDEBUG
      if (m_DataCl->m_DataHSGrp[1 - m_DataCl->m_DataHSToggle].MDataValid == false )
      {
          cout << name() << ": MDataValid is set to Null" << endl;
      } else {
          cout << name() << ": WARNING - putNullMDataValid called and MDataValid is not false (Data Request still in progress?)" << endl;
      }
#endif
  }

  const sc_event& DataHSStartEvent(void) const 
  { return (m_CommCl->DataRequestStartEvent); }

  const sc_event& DataHSEndEvent(void) const 
  { return (m_CommCl->DataRequestEndEvent); }

  void putSDataThreadBusy(unsigned int sdatathreadbusy) 
  { 
    // is the channel in reset?
    if (!get_reset()) {
      // The sdatathreadbusy signal needs to be part of the 
      // channel for this function to work
      assert(m_ParamCl->sdatathreadbusy == true);
      
      m_DataCl->SputSDataThreadBusy( sdatathreadbusy );
    }
  }

  void 
  putNextSDataThreadBusy(unsigned int newValue) 
  { 
    // is the channel in reset?
    if (!get_reset()) {
      // The sdatathreadbusy signal needs to be part of the 
      // channel for this function to work
      assert(m_ParamCl->sdatathreadbusy == true);
      
      m_DataCl->SputNextSDataThreadBusy( newValue );
    }
  }


  unsigned int getSDataThreadBusy(void) const
  { 
    // is the channel in reset?
    if (m_CommCl->Reset) {
      // In reset. Give up.
      return false;
    }
    // The sdatathreadbusy signal must be part of the channel for this command
    // to function correctly
    assert( m_ParamCl->sdatathreadbusy == true );

    return m_DataCl->MgetSDataThreadBusy();
  }

  //---------------------------------------------------------------------
  // public Reset Commands
  //---------------------------------------------------------------------

  bool getReset(void)
  { return (get_reset()); }

  const sc_event& ResetStartEvent(void) const 
  { return (m_CommCl->ResetStartEvent); }

  const sc_event& ResetEndEvent(void) const 
  { return (m_CommCl->ResetEndEvent); }

  void MResetAssert(void)
  { 
      if (! m_DataCl->m_MReset_n) {
          // Master has already caused a reset.
          // Do nothing.
          return;
      }

      assert(m_ParamCl->mreset);

      m_DataCl->m_MReset_n = false;
      m_DataCl->m_MResetStartTime= sc_time_stamp();

      // TODO: This should be a separate function
      // If the channel is not already in reset, reset it.
      if (! get_reset()) {
          // reset the channel
          reset();
      }
  }
      
  void MResetDeassert(void) 
  {
      if (m_DataCl->m_MReset_n) {
          // Master is already out of reset
          // Do nothing.
          return;
      }

      assert(m_ParamCl->mreset);

      m_DataCl->m_MReset_n = true;

      // This command can only be called 16 or more cycles after reset was asserted
      if (m_self_timed)
	assert( (sc_time_stamp() - m_DataCl->m_MResetStartTime) > sc_time(15.99 * m_clkPeriod, m_clkTimeUnit) );

      if (m_DataCl->m_SReset_n) {
          // The slave reset has also been deasserted. Reset is over.
          assert(get_reset());  // Channel should be in reset mode.

          // reset the channel again to make sure all signals are re-initialized
          reset();
          // Now pull out of reset mode
          m_CommCl->Reset = false;
          m_CommCl->ResetEndEvent.notify();
      }
  }

  void SResetAssert(void)
  { 
      if (! m_DataCl->m_SReset_n) {
          // Slave has already caused a reset.
          // Do nothing.
          return;
      }

      assert(m_ParamCl->sreset);

      m_DataCl->m_SReset_n = false;
      m_DataCl->m_SResetStartTime= sc_time_stamp();

      // TODO: This should be a separate function
      // If the channel is not already in reset, reset it.
      if (! get_reset()) {
          // reset the channel
          reset();
      }
  }
      
  void SResetDeassert(void)
  {
      if (m_DataCl->m_SReset_n) {
          // Slave is already out of reset
          // Do nothing.
          return;
      }

      assert(m_ParamCl->sreset);

      m_DataCl->m_SReset_n = true;

      // This command can only be called 16 or more cycles after reset was asserted
      if (m_self_timed)
	assert( (sc_time_stamp() - m_DataCl->m_SResetStartTime) > sc_time(15.99 * m_clkPeriod, m_clkTimeUnit) );

      if (m_DataCl->m_MReset_n) {
          // The master reset has also been deasserted. Reset is over.
          assert(get_reset());  // Channel should be in reset mode.

          // reset the channel again to make sure all signals are re-initialized
          reset();
          // Now pull out of reset mode
          m_CommCl->Reset = false;
          m_CommCl->ResetEndEvent.notify();
      }
  }

  //---------------------------------------------------------------------
  // public OCP Master Sideband Commands
  //---------------------------------------------------------------------
  void MputMError(bool nextValue) 
  { m_DataCl->MputMError(nextValue); }
  void MputMFlag(unsigned int nextValue)
  { m_DataCl->MputMFlag(nextValue); }
  void MputMFlag(unsigned int nextValue, unsigned int mask)
  { m_DataCl->MputMFlag(nextValue, mask); }
  bool MgetSError(void) const
  { return (m_DataCl->MgetSError()); }
  unsigned int MgetSFlag(void) const
  { return (m_DataCl->MgetSFlag()); }
  bool MgetSInterrupt(void) const
  { return (m_DataCl->MgetSInterrupt()); }

  //---------------------------------------------------------------------
  // public OCP Master Sideband Events 
  //---------------------------------------------------------------------
  const sc_event& SidebandSErrorEvent(void) const  
  { return (m_DataCl->SidebandSErrorEvent()); }
  const sc_event& SidebandSFlagEvent(void) const
  { return (m_DataCl->SidebandSFlagEvent()); }
  const sc_event& SidebandSInterruptEvent(void) const
  { return (m_DataCl->SidebandSInterruptEvent()); }

  //---------------------------------------------------------------------
  // public OCP Slave Sideband Commands
  //---------------------------------------------------------------------
  bool SgetMError(void) const
  { return (m_DataCl->SgetMError()); }
  unsigned int SgetMFlag(void) const
  { return (m_DataCl->SgetMFlag()); }
  void SputSError(bool nextValue)
  { m_DataCl->SputSError(nextValue); }
  void SputSFlag(unsigned int nextValue)
  { m_DataCl->SputSFlag(nextValue); }
  void SputSFlag(unsigned int nextValue, unsigned int mask)
  { m_DataCl->SputSFlag(nextValue, mask); }
  void SputSInterrupt(bool nextValue)
  { m_DataCl->SputSInterrupt(nextValue); }

  //---------------------------------------------------------------------
  // public OCP Slave Sideband Events
  //---------------------------------------------------------------------
  const sc_event& SidebandMErrorEvent(void) const
  { return (m_DataCl->SidebandMErrorEvent()); }
  const sc_event& SidebandMFlagEvent(void) const
  { return (m_DataCl->SidebandMFlagEvent()); }

  //---------------------------------------------------------------------
  // public OCP System Sideband Commands
  //---------------------------------------------------------------------
  bool SysputControl(int nextValue)
  { return m_DataCl->SysputControl(nextValue); }
  bool SysgetControlBusy(void) const
  { return (m_DataCl->SysgetControlBusy()); }
  void SysputControlWr(bool nextValue)
  { m_DataCl->SysputControlWr(nextValue); }
  int SysgetStatus(void) const
  { return (m_DataCl->SysgetStatus()); }
  bool readStatus(int& currentValue) const
  { return (m_DataCl->readStatus(currentValue)); }
  bool SysgetStatusBusy(void) const
  { return (m_DataCl->SysgetStatusBusy()); }
  void SysputStatusRd(bool nextValue)
  { m_DataCl->SysputStatusRd(nextValue); }

  //---------------------------------------------------------------------
  // public OCP System Sideband Events
  //---------------------------------------------------------------------
  const sc_event& SidebandControlBusyEvent(void) const
  { return (m_DataCl->SidebandControlBusyEvent()); }
  const sc_event& SidebandStatusEvent(void) const
  { return (m_DataCl->SidebandStatusEvent()); }
  const sc_event& SidebandStatusBusyEvent(void) const
  { return (m_DataCl->SidebandStatusBusyEvent()); }

  //---------------------------------------------------------------------
  // public OCP Core Sideband Commands
  //---------------------------------------------------------------------
  int CgetControl(void) const
  { return (m_DataCl->CgetControl()); }
  void CputControlBusy(bool nextValue)
  { m_DataCl->CputControlBusy(nextValue); }
  bool CgetControlWr(void) const
  { return (m_DataCl->CgetControlWr()); }
  void CputStatus(int nextValue)
  { m_DataCl->CputStatus(nextValue); }
  void CputStatusBusy(bool nextValue)
  { m_DataCl->CputStatusBusy(nextValue); }
  bool CgetStatusRd(void) const
  { return (m_DataCl->CgetStatusRd()); }
                                      
  //---------------------------------------------------------------------
  // public OCP Core Sideband Events
  //---------------------------------------------------------------------
  const sc_event& SidebandControlEvent(void) const 
  { return (m_DataCl->SidebandControlEvent()); }
  const sc_event& SidebandControlWrEvent(void) const
  { return (m_DataCl->SidebandControlWrEvent()); }
  const sc_event& SidebandStatusRdEvent(void) const
  { return (m_DataCl->SidebandStatusRdEvent()); }




  ///////////////////////////////////
  // Port Access to the Generic commands in the underlying Generic Channel
  // Generic commands kept for backward compatibility
  ///////////////////////////////////

  // -----------------------------------------------------------------
  // Generic Direct Channel Access Methods
  // -----------------------------------------------------------------
  virtual bool MputDirect(int MasterID, bool IsWrite, Td *Data, Ta Address,
                          int NumWords)
  {
    return (TL_Channel<TdataCl>::MputDirect(MasterID,IsWrite,Data,Address,NumWords));
  }
  virtual bool SputDirect(int SlaveID, bool IsWrite, Td *Data, Ta Address,int NumWords)
  {
    return (TL_Channel<TdataCl>::SputDirect(SlaveID,IsWrite,Data,Address,NumWords));
  }

  // -----------------------------------------------------------------
  // Direct Access to the internals of the channel.
  // Necessary for the use of the Generic Commands.
  // Note: use with caution.
  // -----------------------------------------------------------------
  TdataCl* GetDataCl(void) 
  { return (TL_Channel<TdataCl>::GetDataCl()); }
  ParamCl<TdataCl>* GetParamCl(void)
  { return (TL_Channel<TdataCl>::GetParamCl()); }
  CommCl* GetCommCl(void) 
  { return (TL_Channel<TdataCl>::GetCommCl()); }

  // -----------------------------------------------------------------
  // Generic Channel Methods
  // -----------------------------------------------------------------
  void Mrelease(sc_time time) { TL_Channel<TdataCl>::Mrelease(time); }
  void Srelease(sc_time time) { TL_Channel<TdataCl>::Srelease(time); }
  void Mrelease(void) { TL_Channel<TdataCl>::Mrelease(); }
  void Srelease(void) { TL_Channel<TdataCl>::Srelease(); }
  void SreleaseData(sc_time time) { TL_Channel<TdataCl>::SreleaseData(time); }
  void SreleaseData(void) { TL_Channel<TdataCl>::SreleaseData(); }
  void MreleasePE(void) { TL_Channel<TdataCl>::MreleasePE(); }
  void MunreleasePE(void) { TL_Channel<TdataCl>::MunreleasePE(); }
  void SreleasePE(void) { TL_Channel<TdataCl>::SreleasePE(); }
  void SunreleasePE(void) { TL_Channel<TdataCl>::SunreleasePE(); }
  void SreleaseDataPE(void) { TL_Channel<TdataCl>::SreleaseDataPE(); }
  void SunreleaseDataPE(void) { TL_Channel<TdataCl>::SunreleaseDataPE(); }
  bool MgetSbusy(void) const { return (TL_Channel<TdataCl>::MgetSbusy()); }
  bool MgetSbusyData(void) const { return (TL_Channel<TdataCl>::MgetSbusyData()); }
  bool SgetMbusy(void) const { return (TL_Channel<TdataCl>::SgetMbusy()); }
  bool IsWrite(void) { return (TL_Channel<TdataCl>::IsWrite()); }
  void SregisterDirectIF(SdirectIF<TdataCl>* A)
  { TL_Channel<TdataCl>::SregisterDirectIF(A); }
  void MregisterDirectIF(MdirectIF<TdataCl>* A)
  { TL_Channel<TdataCl>::MregisterDirectIF(A); }
  bool MputWriteRequest(void)
  { return (TL_Channel<TdataCl>::MputWriteRequest()); }
  bool MputReadRequest(void)
  { return (TL_Channel<TdataCl>::MputReadRequest()); }
  bool MputWriteRequestBlocking(void) 
  { return (TL_Channel<TdataCl>::MputWriteRequestBlocking()); }
  bool MputReadRequestBlocking(void) 
  { return (TL_Channel<TdataCl>::MputReadRequestBlocking()); }
  bool MputRequest(void) { return (TL_Channel<TdataCl>::MputRequest()); }
  bool MputRequestBlocking(void)
  { return (TL_Channel<TdataCl>::MputRequestBlocking()); }
  bool SgetRequest(bool Release)
  { return (TL_Channel<TdataCl>::SgetRequest(Release)); }
  bool SgetRequestBlocking(bool Release)
  { return (TL_Channel<TdataCl>::SgetRequestBlocking(Release)); }
  bool SgetRequestPE(void) { return (TL_Channel<TdataCl>::SgetRequestPE()); }
  bool SputResponse(void) { return (TL_Channel<TdataCl>::SputResponse()); }
  bool SputResponseBlocking(void)
  { return (TL_Channel<TdataCl>::SputResponseBlocking()); }
  bool MgetResponse(bool Release)
  { return (TL_Channel<TdataCl>::MgetResponse(Release)); }
  bool MgetResponseBlocking(bool Release)
  { return (TL_Channel<TdataCl>::MgetResponseBlocking(Release)); }
  bool MgetResponsePE(void)
  { return (TL_Channel<TdataCl>::MgetResponsePE()); }
  bool MputDataRequest(void)
  { return (TL_Channel<TdataCl>::MputDataRequest()); }
  bool MputDataRequestBlocking(void)
  { return (TL_Channel<TdataCl>::MputDataRequestBlocking()); }
  bool SgetDataRequest(bool Release)
  { return (TL_Channel<TdataCl>::SgetDataRequest(Release)); }
  bool SgetDataRequestBlocking(bool Release)
  { return (TL_Channel<TdataCl>::SgetDataRequestBlocking(Release)); }
  bool SgetDataRequestPE(void)
  { return (TL_Channel<TdataCl>::SgetDataRequestPE()); }

  void reset()
  { TL_Channel<TdataCl>::reset(); }
  void remove_reset()
  { TL_Channel<TdataCl>::remove_reset(); }
  bool get_reset()
  { return (TL_Channel<TdataCl>::get_reset()); }


  // -----------------------------------------------------------------
  // Generic Channel Events
  // -----------------------------------------------------------------
  const sc_event& default_event() const
  { return (TL_Channel<TdataCl>::default_event()); }


protected:
  //---------------------------------------------------------------
  // protected methods
  //---------------------------------------------------------------

  // display a fatal error message and quick the program
  void fatalError(char* sub_title, char* msg)
  {
    cout << endl << name() << ": ERROR: " << sub_title << ":" << msg << endl;
    assert(0);
  }

  //----------------------------------------------------------------------
  // member data
  //----------------------------------------------------------------------

  // OCP clock information
  double       m_clkPeriod;
  sc_time_unit m_clkTimeUnit;

  bool         m_doRunCheck;

  // OCP Trace variables
  bool m_doOCPMon;


  // Address Mask - used to make sure user address is not to big for the channel.
  Ta m_addressMask;  

  // Data Mask - used to make sure user data is not to big for the channel.
  Td m_dataMask;  

  // Flag - set if a fatal error has occured and the simulation should be stopped
  // once the last line of the OCP Trace File has been output (for debugging).
  bool m_exitAfterOCPMon;

  // Are blocking calls clocked or self timed?
  bool m_clocked;
  bool m_self_timed; // sometimes can be not self timed and not clocked
  sc_clock * m_clk;
  sc_in_clk * m_in_clk;
  sc_signal<bool> * m_sig_clk;
  
  // OCP monitor
  OCPMon< OCP_TL1_DataCl<Td, Ta> > * m_monitor;
};

#endif  // _OCP_TL1_CHANNEL_H
