// 
//  Copyright 2003 OCP-IP
//  OCP-IP Confidential & Proprietary
//
// ============================================================================
//      Project : OCP SLD WG
//      Authors : Yann Bajot, PROSILOG, bajot@prosilog.com 
//                Alan Kamas, Sonics Inc., aok@sonicsinc.com 
//         Date : 10/16/2003
//
//  Description:  Regresion test 3 for the OCP-specific TL2 API.
//                Tests the behaviour of the master/slave interface methods
//                described in the OCP TL2 specific channel, in a
//                multi-threaded context. The simulation behaviour is the same
//                as for the 'complex_example' located in the 'examples/tl2'
//                directory. See section 7 of the OCP API documentation for
//                more details.
//
// ============================================================================


///////////////////////////
//
// Simple Main to read in Map data from files
// and then use that to configure and connect
// a master and slave.
//
///////////////////////////



#include <map>
#include <set>
#include <string>
#include <algorithm>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

#include "systemc.h"

#include "MasterSysC.h"
#include "SlaveSysC.h"
#include "ocp_tl2_data_cl.h"
#include "ocp_tl_param_cl.h"
#include "ocp_tl2_channel.h"

void process_command_line(int   argc,
                          char* argv[],
                          string& ocp_params_file_name,
                          string& master_params_file_name,
                          string& slave_params_file_name,
                          double& simulation_end_time,
                          bool& debug_dump,
                          string& debug_file_name)
{
    // get the ocp parameters file name
    ocp_params_file_name = "";
    if (argc > 1) {
        string file_name(argv[1]);
        ocp_params_file_name = file_name;
    }

    // get the master parameters file name
    master_params_file_name = "";
    if (argc > 2) {
        string file_name(argv[2]);
        master_params_file_name = file_name;
    }
    // get the slave parameters file name
    slave_params_file_name = "";
    if (argc > 3) {
        string file_name(argv[3]);
        slave_params_file_name = file_name;
    }

    // get the simulation end time
    simulation_end_time = 1000;
    if (argc > 4) {
        simulation_end_time = (double) atol(argv[4]);
    }

    // do we dump out a log file?
    debug_dump= false;
    debug_file_name = "";
    if (argc > 5) {
        string file_name(argv[5]);
        debug_file_name = file_name;
        debug_dump = true;
    }
}

void readMapFromFile(const string &myFileName, MapStringType &myParamMap) 
{
    // read pairs of data from the passed file
    string leftside;
    string rightside;
    
    // (1) open the file
    ifstream inputfile(myFileName.c_str());
    assert( inputfile );

    // set the formatting
    inputfile.setf(std::ios::skipws);

    // Now read through all the pairs of values and add them to the passed map
    while ( inputfile ) {
        inputfile >> leftside;
        inputfile >> rightside;
        myParamMap.insert(std::make_pair(leftside,rightside));
    }

    // All done, close up
    inputfile.close();
}

int
sc_main(int argc, char* argv[])
{
    OCP_TL2_Channel< unsigned int, unsigned int >* pOCP;
    MasterSysC*          pMaster;
    SlaveSysC*           pSlave;
    MapStringType                                                ocpParamMap;
    MapStringType                                                masterParamMap;
    MapStringType                                                slaveParamMap;

    double         simulation_end_time;
    bool           debug_dump;
    string         ocpParamFileName;
    string         masterParamFileName;
    string         slaveParamFileName;
    string         dump_file_name;
    ofstream       debugFile;

    // --------------------------------
    // (1) process command line options
    //     and read my parameters
    // --------------------------------
    process_command_line(argc,argv,ocpParamFileName,masterParamFileName,
           slaveParamFileName,simulation_end_time,debug_dump,dump_file_name);

    if ( ! ocpParamFileName.empty() ) {
        readMapFromFile(ocpParamFileName, ocpParamMap);
    }

    if ( ! masterParamFileName.empty() ) {
        readMapFromFile(masterParamFileName, masterParamMap);
    }

    if ( ! slaveParamFileName.empty() ) {
        readMapFromFile(slaveParamFileName, slaveParamMap);
    }

    // open a trace file
    if (debug_dump) {
        cout << "Debug dumpfilename: " << dump_file_name << endl;
        debugFile.open(dump_file_name.c_str());
    }

    // Default time unit is SC_NS
    sc_set_default_time_unit(1,SC_NS);

    // ----------------------------------------------------------
    // (2) Create the OCP Channel
    // ----------------------------------------------------------

    pOCP = new OCP_TL2_Channel<unsigned int, unsigned int>
        ("ocp0");
    pOCP->setConfiguration(ocpParamMap);

    // ----------------------------------------------------------
    // (3) Create the Master and Slave
    // ----------------------------------------------------------

    pMaster = new MasterSysC("master",  &debugFile );
    pMaster->setConfiguration(masterParamMap);

    pSlave = new SlaveSysC("slave", 0x3FF, &debugFile );
    pSlave->setConfiguration(slaveParamMap);

    // ----------------------------------------------------------
    // (4) connect channel, master, and slave, & clock
    // ----------------------------------------------------------
    pMaster->ipP(*pOCP);
    pSlave->tpP(*pOCP);

    // ------------------------
    // (5) start the simulation
    // ------------------------
    sc_start(simulation_end_time,SC_NS);

    // -------------------
    // (6) post processing
    // -------------------

    /*
    cout << "main program finished at " 
         << sc_time_stamp().to_double() << endl;

    sc_simcontext* sc_curr_simcontext = sc_get_curr_simcontext();
    cout << "delta_count: " << dec << sc_curr_simcontext->delta_count() << endl;
    cout << "next_proc_id: " << dec << sc_curr_simcontext->next_proc_id()
         << endl;
    */
    
    return (0);
}
